/***************************************************************************
                          tea_config.c  -  description
                             -------------------
    begin                : Fir Dec 19 2003
    copyright            : (C) 2003 by Peter 'Roxton' Semiletov
    email                : tea@list.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <errno.h>
#include <sys/stat.h>

#include <glib.h>
#include <time.h>
#include <stdio.h>

#include "callbacks.h"
#include "interface.h"
#include "tea_defs.h"
#include "tea_text_document.h"
#include "tea_config.h"
#include "tea_hl.h"
#include "tea_spell.h"


static GHashTable *ht_tea_hotkeys = NULL;
static GHashTable *tea_options = NULL;


void ui_init (void)
{
  GList *ui = NULL;
  gl_famous = NULL;
  gl_kwas_history = NULL;
  gl_tea_kwas_bookmarks = NULL;
 
  if (g_file_test (confile.tea_ui, G_FILE_TEST_EXISTS))
      ui = load_file_to_glist (confile.tea_ui);

  if (g_file_test (confile.tea_gl_famous, G_FILE_TEST_EXISTS))
      gl_famous = load_file_to_glist (confile.tea_gl_famous);

  if (g_file_test (confile.tea_kwas_history, G_FILE_TEST_EXISTS))
      gl_kwas_history = load_file_to_glist (confile.tea_kwas_history);

  if (g_file_test (confile.tea_kwas_bookmarks, G_FILE_TEST_EXISTS))
      gl_tea_kwas_bookmarks = load_file_to_glist (confile.tea_kwas_bookmarks);


  confile.last_prefs_tab = conf_get_int_value (ui, "last_prefs_tab", 0); 

  gtk_paned_set_position (vpaned1, conf_get_int_value (ui, "vpaned1_y", 300));
                  
  gtk_window_move (GTK_WINDOW (tea_main_window),
                   conf_get_int_value (ui, "main_wnd_x", 2),
                   conf_get_int_value (ui, "main_wnd_y", 2));

  gtk_window_resize (GTK_WINDOW (tea_main_window),
                     conf_get_int_value (ui, "main_wnd_w", 630),
                     conf_get_int_value (ui, "main_wnd_h", 450));

  if (ui)
     glist_strings_free (ui);
}


void ui_done (void)
{
  GList *list = NULL;

  list = add_to_glist_combined_int (list, "vpaned1_y", gtk_paned_get_position  (vpaned1));

  gint x;
  gint y;
  
  gtk_window_get_position (GTK_WINDOW (tea_main_window), &x, &y);

  list = add_to_glist_combined_int (list, "last_prefs_tab", confile.last_prefs_tab);
  list = add_to_glist_combined_int (list, "main_wnd_x", x);
  list = add_to_glist_combined_int (list, "main_wnd_y", y);

  gtk_window_get_size (GTK_WINDOW (tea_main_window), &x, &y);

  list = add_to_glist_combined_int (list, "main_wnd_w", x);
  list = add_to_glist_combined_int (list, "main_wnd_h", y);
 
  glist_save_to_file (list, confile.tea_ui);
  glist_save_to_file (gl_famous, confile.tea_gl_famous);
  glist_save_to_file (gl_kwas_history, confile.tea_kwas_history);
  glist_save_to_file (gl_tea_kwas_bookmarks, confile.tea_kwas_bookmarks);

  g_list_free (list);
  glist_strings_free (gl_famous);
  glist_strings_free (gl_kwas_history);
  glist_strings_free (gl_tea_kwas_bookmarks);
}


void confile_reload (void)
{
  confile.confdata = NULL;

/*  GdkScreen *user_screen = gdk_screen_get_default ();

  if (user_screen)
     {
      confile.screen_h = gdk_screen_get_height (user_screen); 
      confile.screen_w = gdk_screen_get_width (user_screen);
     }
  else
      {
       confile.screen_h = 480;
       confile.screen_w = 640;
      }
   */

  confile.screen_w = gdk_screen_width ();
  confile.screen_h = gdk_screen_height ();

  gchar *s = g_get_home_dir ();
  confile.user_tea_main_dir = g_strconcat (s, G_DIR_SEPARATOR_S, ".tea", NULL);

  confile.tmp_dir = g_strdup (g_get_tmp_dir ());
  confile.tmp_file = g_strconcat (confile.tmp_dir, G_DIR_SEPARATOR_S, "tea_temp", NULL);
  confile.user_menu_file = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "tea_user_menu", NULL); 

  confile.bmx_file = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "tea_bmx", NULL); 

  confile.sessions_dir = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "sessions", G_DIR_SEPARATOR_S, NULL);
  confile.snippets_dir = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "snippets", G_DIR_SEPARATOR_S, NULL);
  confile.user_tea_templates_dir = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "templates", G_DIR_SEPARATOR_S, NULL);
  confile.iconv_file = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "iconv_rc", NULL);
  confile.crapbook_file = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "crapbook.txt", NULL);

  confile.autocomp_file = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "autocomp_file", NULL);

  confile.tea_rc = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "tea_rc", NULL);
  confile.tea_ui = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "tea_ui", NULL);
  confile.tea_gl_famous = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "tea_gl_famous", NULL);
  confile.tea_kwas_history = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "tea_kwas_history", NULL);
  confile.tea_kwas_bookmarks = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "tea_kwas_bookmarks", NULL);

  confile.tea_recent = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "tea_recent", NULL);
  confile.tea_hotkeys = g_strconcat (confile.user_tea_main_dir, G_DIR_SEPARATOR_S, "tea_hotkeys", NULL);
  
  //FIXME: rewrite all here
  if (! g_file_test (confile.user_tea_main_dir, G_FILE_TEST_IS_DIR))
      //create tea directories
      {
       chdir (s);
       if (mkdir (".tea", S_IRUSR | S_IWUSR | S_IXUSR) == -1) 
           DBM ("mkdir tea_dir failed");
       
       chdir (".tea");

       if (mkdir ("templates", S_IRUSR | S_IWUSR | S_IXUSR) == -1)
           DBM ("mkdir templates failed");

       if (mkdir ("sessions", S_IRUSR | S_IWUSR | S_IXUSR) == -1)
           DBM ("mkdir sessions failed");

       chdir (s);
      }


  if (! g_file_test (confile.tea_rc, G_FILE_TEST_EXISTS))
      create_empty_file (confile.tea_rc, "#TEA editor configuration file. Please read the manual");

  if (! g_file_test (confile.tea_hotkeys, G_FILE_TEST_EXISTS))
      create_empty_file (confile.tea_hotkeys, "#TEA hotkeys file. It MUST be in UTF-8! Please read the manual");

  confile.confdata = parse_conf_file_to_list (confile.tea_rc);

  confile.editor_font = conf_get_char_value (confile.confdata, "editor_font", "Monospace 14");

  confile.ext_pic_editor = conf_get_char_value (confile.confdata, "ext_pic_editor", "gimp %s&");
  confile.ext_pic_viewer = conf_get_char_value (confile.confdata, "ext_pic_viewer", "display %s&");
  confile.date_time = conf_get_char_value (confile.confdata, "date_time", "%d/%m/%Y %T");
  confile.max_recent_items = conf_get_int_value (confile.confdata, "max_recent_items", 7);
  confile.famous_history_max = (conf_get_int_value (confile.confdata, "famous_history_max", 7) + 1);

  confile.saveopen_h = conf_get_int_value (confile.confdata, "saveopen_h", 75);
  confile.saveopen_w = conf_get_int_value (confile.confdata, "saveopen_w", 75);
  
  confile.show_line_nums = conf_get_int_value (confile.confdata, "show_line_nums", 0);
  confile.word_wrap = conf_get_int_value (confile.confdata, "word_wrap", 1);
  confile.tab_size = conf_get_int_value (confile.confdata, "tab_size", 3);

  confile.text_foreground = conf_get_char_value (confile.confdata, "text_foreground", "black");
  confile.text_background = conf_get_char_value (confile.confdata, "text_background", "white");
  confile.text_sel_foreground = conf_get_char_value (confile.confdata, "text_sel_foreground", "white");
  confile.text_sel_background = conf_get_char_value (confile.confdata, "text_sel_background", "black");
 
  confile.tag_comment = conf_get_char_value (confile.confdata, "tag_comment", "#605454");
  confile.tag_identifier = conf_get_char_value (confile.confdata, "tag_identifier", "black");
  confile.tag_digit = conf_get_char_value (confile.confdata, "tag_digit", "navy");
  confile.tag_string = conf_get_char_value (confile.confdata, "tag_string", "red");
  confile.tag_html_tag = conf_get_char_value (confile.confdata, "tag_html_tag", "black");
  confile.tag_preprocessor = conf_get_char_value (confile.confdata, "tag_preprocessor", "#b522a6");
  confile.tag_type = conf_get_char_value (confile.confdata, "tag_type", "#3650bd");
  confile.tag_spellcheck = conf_get_char_value (confile.confdata, "tag_spellcheck", "red"); //#e70f1e

  confile.tag_comment_font = conf_get_char_value (confile.confdata, "tag_comment_font", "Monospace Italic 14");
  confile.tag_identifier_font = conf_get_char_value (confile.confdata, "tag_identifier_font", "Monospace Bold 14");
  confile.tag_digit_font = conf_get_char_value (confile.confdata, "tag_digit_font", confile.editor_font);
  confile.tag_string_font = conf_get_char_value (confile.confdata, "tag_string_font", confile.editor_font);
  confile.tag_html_tag_font = conf_get_char_value (confile.confdata, "tag_html_tag_font", "Monospace Bold 14");
  confile.tag_preprocessor_font = conf_get_char_value (confile.confdata, "tag_preprocessor_font", confile.editor_font);
  confile.tag_type_font = conf_get_char_value (confile.confdata, "tag_type_font", confile.editor_font);


  confile.cm_mplayer = conf_get_char_value (confile.confdata, "cm_mplayer", "mplayer \"@movie_file\" -subcp @encoding -ss @time -sub \"@sub_file\" &");

  //n.p. Scorn - Orgy Of Holyness

  //some default values were taken from Bluefish

  confile.cmd_Opera = conf_get_char_value (confile.confdata, "opera","opera -remote 'openURL(%s,new-window)' || opera %s&");
  confile.cmd_Firefox = conf_get_char_value (confile.confdata, "firefox", "firefox %s&");
  confile.cmd_Mozilla = conf_get_char_value (confile.confdata, "mozilla", "mozilla -remote 'openURL(%s, new-window)' || mozilla %s&");
  confile.cmd_Konqueror = conf_get_char_value (confile.confdata, "konqueror", "konqueror %s&");
  confile.cmd_Galeon = conf_get_char_value (confile.confdata, "galeon", "galeon -x %s&");
  confile.cmd_Dillo = conf_get_char_value (confile.confdata, "dillo", "dillo %s&");
  confile.cmd_Links = conf_get_char_value (confile.confdata, "links", "konsole -e links %s&");
  confile.cmd_ELinks = conf_get_char_value (confile.confdata, "elinks", "konsole -e elinks %s&");
  
  confile.cmd_Lynx = conf_get_char_value (confile.confdata, "lynx", "konsole -e lynx %s&");
  confile.cmd_Epiphany = conf_get_char_value (confile.confdata, "epiphany", "epiphany %s&");
  confile.cmd_Other = conf_get_char_value (confile.confdata, "other_browser", "%s&");
  confile.cmd_def_doc_browser = conf_get_char_value (confile.confdata, "def_doc_browser", "mozilla %s&");
  confile.use_def_doc_browser = conf_get_int_value (confile.confdata, "use_def_doc_browser", 0);

  confile.rtf_enc = conf_get_char_value (confile.confdata, "rtf_enc", "CP1251");
  
  confile.main_wnd_show_full_path = conf_get_int_value (confile.confdata, "main_wnd_show_full_path", 1);
  confile.do_backup = conf_get_int_value (confile.confdata, "do_backup", 0);
  confile.do_hl_on_fileopen = conf_get_int_value (confile.confdata, "do_hl_on_fileopen", 1);

  confile.kwas_show_hidden_files = conf_get_int_value (confile.confdata, "kwas_show_hidden_files", 1);

  confile.use_snippets = conf_get_int_value (confile.confdata, "use_snippets", 1);
  confile.use_ext_image_viewer = conf_get_int_value (confile.confdata, "use_ext_image_viewer", 0);

  confile.do_enc_autodetect = conf_get_int_value (confile.confdata, "do_enc_autodetect", 0);

  confile.enc_det_cp1251 = conf_get_int_value (confile.confdata, "enc_det_cp1251", 0); 
  confile.enc_det_koi8r = conf_get_int_value (confile.confdata, "enc_det_koi8r", 0);
  confile.enc_det_koi8u = conf_get_int_value (confile.confdata, "enc_det_koi8u", 0);
  confile.enc_det_866 = conf_get_int_value (confile.confdata, "enc_det_866", 0);
  confile.enc_det_finnish = conf_get_int_value (confile.confdata, "enc_det_finnish", 0);
  confile.enc_det_german = conf_get_int_value (confile.confdata, "enc_det_german", 0);
  confile.enc_det_serbian = conf_get_int_value (confile.confdata, "enc_det_serbian", 0);
  confile.enc_det_latvian = conf_get_int_value (confile.confdata, "enc_det_latvian", 0);
  confile.enc_det_polish = conf_get_int_value (confile.confdata, "enc_det_polish", 0);
  confile.enc_det_portuguese = conf_get_int_value (confile.confdata, "enc_det_portuguese", 0);
  confile.enc_det_slovak = conf_get_int_value (confile.confdata, "enc_det_slovak", 0);
  confile.enc_det_slovenian = conf_get_int_value (confile.confdata, "enc_det_slovenian", 0);
  confile.enc_det_spanish = conf_get_int_value (confile.confdata, "enc_det_spanish", 0);
  confile.enc_det_turkish = conf_get_int_value (confile.confdata, "enc_det_turkish", 0);

  confile.scan_for_links_on_doc_open = conf_get_int_value(confile.confdata, "scan_for_links_on_doc_open", 0);
  confile.logmemo_lines_max = conf_get_int_value(confile.confdata, "logmemo_lines_max", 13);

  get_iconv_sup();
}


GList* parse_conf_file_to_list (gchar *filename)
{
  gsize length;
  gchar *buf;
  GList *result = NULL;

  if  (! g_file_get_contents (filename, &buf, &length, NULL))
       return NULL;

  result = parse_string_to_list (buf);
  g_free (buf);
  return result;
}


gchar* conf_get_char_value (GList *confdata, gchar *key, gchar *def)
{
  if (! confdata)
     return g_strdup (def);

  gchar *result = NULL;
  gchar **cms = NULL;

  GList *p = g_list_first (confdata);

  while (p)
        {
         cms = g_strsplit ((gchar*) p->data, "=", 2);

         if (cms)
           {
            if (cms[0])
               if ((g_ascii_strncasecmp (cms[0], key, 512) == 0))
                  {
                   if (cms[1])
                      {
                       if (g_ascii_strncasecmp (cms[1], "", 512) == 0)
                          result = strdup (def);
                       else
                           result = strdup (cms[1]);
                      }
                   g_strfreev (cms);
                   return result;
                  }

            g_strfreev (cms);
            p = g_list_next (p);
           }
        }

  if (! def)
     return NULL;

  return strdup (def);
}  


gint conf_get_int_value (GList *confdata, gchar *key, gint def)
{
  if (! confdata)
    return def;

  gint result = def;                                 
             
  gchar *s = g_strdup_printf ("%d", def);  
  gchar *res = conf_get_char_value (confdata, key, s);

  if (! res)
     {
      g_free (s);
      return def;
     }  
  
  result = strtol (res, NULL, 10);

  g_free (res);
  g_free (s);

  return result;
}


//FIXME: need for cleanup
void confile_free (void)
{
  glist_strings_free (confile.confdata);

  g_free (confile.tea_kwas_history);
  g_free (confile.tea_kwas_bookmarks);
  g_free (confile.user_menu_file);
  g_free (confile.crapbook_file);
  g_free (confile.snippets_dir);
  g_free (confile.bmx_file);
  g_free (confile.date_time);                                      
  g_free (confile.user_tea_main_dir);
  g_free (confile.user_tea_templates_dir);
  g_free (confile.tmp_dir);
  g_free (confile.tmp_file);
  g_free (confile.iconv_file);
  g_free (confile.autocomp_file);
  g_free (confile.sessions_dir);
  g_free (confile.tea_rc);
  g_free (confile.tea_ui);
  g_free (confile.tea_gl_famous);

  g_free (confile.tea_recent);
  g_free (confile.tea_hotkeys);


  g_free (confile.cmd_def_doc_browser);
  g_free (confile.cmd_Opera);
  g_free (confile.cmd_Firefox);
  g_free (confile.cmd_Mozilla);
  g_free (confile.cmd_Konqueror);
  g_free (confile.cmd_Galeon);
  g_free (confile.cmd_Dillo);
  g_free (confile.cmd_Links);
  g_free (confile.cmd_ELinks);
  g_free (confile.cmd_Lynx);
  g_free (confile.cmd_Epiphany);
  g_free (confile.cmd_Other);

  g_free (confile.text_foreground);
  g_free (confile.text_background);
  g_free (confile.text_sel_foreground);
  g_free (confile.text_sel_background);

  g_free (confile.tag_comment);
  g_free (confile.tag_identifier);
  g_free (confile.tag_digit);
  g_free (confile.tag_string);
  g_free (confile.tag_html_tag);
  g_free (confile.tag_preprocessor);
  g_free (confile.tag_type);
  g_free (confile.tag_spellcheck);

  g_free (confile.tag_comment_font);
  g_free (confile.tag_identifier_font);
  g_free (confile.tag_digit_font);
  g_free (confile.tag_string_font);
  g_free (confile.tag_html_tag_font);
  g_free (confile.tag_preprocessor_font);
  g_free (confile.tag_type_font);

  g_free (confile.editor_font);
  g_free (confile.rtf_enc);
  g_free (confile.cm_mplayer);
}


void get_iconv_sup (void)
{
  gchar *cmd = g_strconcat ("iconv -l > ", confile.iconv_file, NULL);
     
  if (! g_file_test (confile.iconv_file, G_FILE_TEST_EXISTS))
     system (cmd);
     
  gsize length;
  gchar *buf;

  if (! g_file_get_contents (confile.iconv_file, &buf, &length, NULL))
      return;

  if (confile.iconv_encs)
      g_list_free (confile.iconv_encs);

  confile.iconv_encs = NULL;
  confile.iconv_encs = parse_sep_string_to_list (buf, "//");
  
  confile.iconv_encs = g_list_prepend (confile.iconv_encs, g_strdup (CURR_LOCALE));
  confile.iconv_encs = g_list_prepend (confile.iconv_encs, g_strdup ("UTF-8"));
  
  if (confile.do_enc_autodetect) 
     confile.iconv_encs = g_list_prepend (confile.iconv_encs, g_strdup ("autodetect"));

  g_free (cur_settings.selected_enc);  
  cur_settings.selected_enc = g_strdup ("UTF-8");

  g_free (cmd);
  g_free (buf);
}

//FIXME?
GList* add_recent_item_composed (GList *list, t_note_page *doc)
{
  if (! doc) return list;

  GList *l = list;
  //l = del_if_strstr (l, doc->file_name); 
 
  l = add_to_list_with_limit2 (list, g_strdup_printf ("%s,%s,%d",
                               doc->file_name, doc->encoding, 
                               editor_get_pos (doc)),
                               confile.max_recent_items);
  return l;
}


void execute_recent_item (gchar *item)
{
  if (! item) return;
  
  if (g_utf8_strlen (item, -1) <= 2)
     return;

  gchar **a = NULL;
  gchar *f = NULL; 

  a = g_strsplit (item, ",", -1);

  if (! a)
     return;

  if (! a[0])
     return;

  if (! a[1])
     return;

  if (! a[2])
     return;

  guint pos = strtol (a[2], NULL, 10);

  f = get_l_filename (a[0]);

  if (! g_file_test ( f, G_FILE_TEST_EXISTS))
     {
      log_to_memo (_("%s is not exists"), f, LM_ERROR);
      g_strfreev (a);
      return;
     }

  g_free (cur_settings.selected_enc);
  cur_settings.selected_enc = g_strdup (a[1]);
  cur_text_doc = doc_open_file (f);

  if (cur_text_doc)
     {
      set_title (cur_text_doc);
      editor_set_pos (cur_text_doc, pos);
     }
  
  g_strfreev (a);
  g_free (f);
}


void update_enc_menu (void)
{
  gtk_widget_destroy (mni_co_menu);
  mni_co_menu = new_menu_submenu (mni_co);

  GList *p = g_list_first (confile.iconv_encs);

  while (p)
       {
        mni_temp = new_menu_item ((gchar*) p->data, mni_co_menu, on_mni_co_select);
        p = g_list_next (p);
       }
}


void bmx_reload (void)
{
  glist_strings_free (bookmarks);
 
  bookmarks = load_file_to_glist (confile.bmx_file);

  if (! bookmarks)
     return;

  gtk_widget_destroy (mni_bookmarks_menu);
  mni_bookmarks_menu = new_menu_submenu (mni_bookmarks_root);

  GList *p = g_list_first (bookmarks);
 
  while (p)
        {
         if ( strcmp (p->data, "") != 0 )
             mni_temp = new_menu_item ((gchar*)p->data, mni_bookmarks_menu, on_mni_recent_activate);
        
         p = g_list_next(p);
        }
}


void update_recent_list_menu (gboolean load_from_file)
{
  if (! mni_recent_list)
     return;

  if (load_from_file)
      {
       glist_strings_free (recent_list);
       recent_list = load_file_to_glist_with_limit (confile.tea_recent, confile.max_recent_items); 
      }

  if (! recent_list)
     return;

  gtk_widget_destroy (mni_recent_menu);
  mni_recent_menu = new_menu_submenu (mni_recent_list);

  gint c = -1;
  GList *p = g_list_first (recent_list);

  while (p)
        {
         if (++c == confile.max_recent_items)
            break;

         if (p->data)
            if ( strcmp (p->data, "") != 0 )
               mni_temp = new_menu_item ((gchar*)p->data, mni_recent_menu, on_mni_real_recent_activate);
         p = g_list_next (p);
        }
}

//FIXME?
void add_recent_internal (t_note_page *doc)
{
  /*if (g_list_length (recent_list) > confile.max_recent_items)
    {
     GList *p = g_list_last (recent_list);
     g_free (p->data);
     recent_list = g_list_delete_link (recent_list, p);
    }
*/
  recent_list = add_recent_item_composed (recent_list, doc);
}


void tea_start (void)
{
  sr_wnd = NULL;
  ht_user_menu = NULL;
  dox = NULL;
  bookmarks = NULL;
  confile.msg_counter = 0;
  confile.iconv_encs = NULL;
  confile.current_movie = NULL;
  cur_settings.selected_enc = NULL;
  if (confile.do_enc_autodetect)
       cur_settings.selected_enc = g_strdup ("autodetect");
  else
      cur_settings.selected_enc = g_strdup ("UTF-8");
}

//n.p. Angelo Badalamenti - Red Bats With Teeth


void tea_done (void)
{
  t_note_page *page;
 
  mni_tabs = NULL;

  GList *p = g_list_first (dox);
  while (p)
        {
         page = (t_note_page*) p->data;
         page_free (page);
         page->text_buffer = NULL;
         p = g_list_next (p);
        }

  glist_save_to_file (recent_list, confile.tea_recent);

  g_list_free (dox);
  glist_strings_free (recent_list);
  glist_strings_free (bookmarks);

  if (ht_tea_hotkeys)
     g_hash_table_destroy (ht_tea_hotkeys); 
  if (ht_user_menu)
     g_hash_table_destroy (ht_user_menu); 

  g_hash_table_destroy (c_t_keywords); 
  g_hash_table_destroy (c_t_types); 
  g_hash_table_destroy (pas_t_keywords); 
  g_hash_table_destroy (pas_t_types); 

  unlink (confile.tmp_file);

  confile_free();
}


GdkModifierType string_to_mod (const gchar *s)
{
  if (! s)
     return 0;

  GdkModifierType result = 0;

  if (strstr (s, "SHIFT") != 0)
     result = result | GDK_SHIFT_MASK;
  
  if (strstr (s, "CTRL") != 0)
     result = result | GDK_CONTROL_MASK;

  if (strstr (s, "ALT") != 0)
     result = result | GDK_MOD1_MASK;
  
  return result;
}


void lookup_widget_cb (GtkWidget *widget, gpointer data)
{
  if (! widget)
     return;

  if (! ht_tea_hotkeys)
      return;
  
  gchar *value = NULL;
  gchar *key;
  gchar *t;
  gchar *s = gtk_widget_get_name (widget);

  if (strcmp ("-x-", s) != 0)
     {
      value = g_hash_table_lookup (ht_tea_hotkeys, s);

      if (value)
         {
          t = g_strrstr (value, " "); 
          if (! t)
             key = g_strdup (value);
          else 
              key = g_strchug (g_strdup (t));
             

          gtk_widget_add_accelerator (widget, "activate", accel_group,
                                      gdk_keyval_from_name (key), string_to_mod (value),
                                      GTK_ACCEL_VISIBLE);
  
          g_free (key); 
         }

      if (GTK_IS_MENU_ITEM (widget))
         lookup_widget (gtk_menu_item_get_submenu (widget), data);      
     }   
}


GtkWidget* lookup_widget (GtkContainer *widget, gchar *widget_name)
{
  if (widget)
     gtk_container_foreach (widget, lookup_widget_cb, widget_name);

  return NULL;
}


void reload_hotkeys (void)
{
  if (ht_tea_hotkeys)
     g_hash_table_destroy (ht_tea_hotkeys); 
  
  gchar *shortcut = NULL;
  gchar *menuitem = NULL;

  ht_tea_hotkeys = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);


  GList *list = parse_conf_file_to_list (confile.tea_hotkeys);
  
  gchar **a = NULL;

  GList *p = g_list_first (list);

  while (p)
        {
         a = g_strsplit ((gchar*) p->data, "=", 2);

         if (a[1])
            shortcut = g_ascii_strup (a[1], -1);

         if (a[0])
            menuitem = g_strdup (a[0]);
 
         if ( shortcut && menuitem)
            g_hash_table_insert (ht_tea_hotkeys, menuitem, shortcut);  

         g_strfreev (a);
         p = g_list_next (p); 
        }

        glist_strings_free (list);
  
  lookup_widget (GTK_CONTAINER (menubar1), "test");
}


void reload_snippets (void)
{
  if (! confile.use_snippets)
     return;
 
  if (! g_file_test (confile.snippets_dir, G_FILE_TEST_IS_DIR))
     if ( mkdir (confile.snippets_dir, S_IRUSR | S_IWUSR | S_IXUSR) == -1)
        return;
 
  gtk_widget_destroy (mni_snippets_menu);
  mni_snippets_menu = new_menu_submenu (mni_snippets);
  GtkWidget *temp = new_menu_tof (mni_snippets_menu);

  GDir *d = g_dir_open (confile.snippets_dir, 0, NULL);  
  gchar *t;

  while (t = g_dir_read_name (d))
        {
         if (! g_file_test (t, G_FILE_TEST_IS_DIR))
             mni_temp = new_menu_item (t, mni_snippets_menu, on_mni_snippet_click);
        }

  g_dir_close (d);  
}


void tabs_reload (void)
{
  if (! mni_tabs)
     return;

  gtk_widget_destroy (mni_tabs_menu);
  mni_tabs_menu = new_menu_submenu (mni_tabs);

  GList *p = g_list_first (dox);
  t_note_page *doc;

  while (p)
       {
        doc = p->data;
        if (doc)                 
            mni_temp = new_menu_item_with_udata (doc->file_name, mni_tabs_menu, on_mni_tabs_menuitem_click, doc);
        
        p = g_list_next (p);
       }
}

//n.p.Trio Bayanistov - Barynja (1937)


void reload_dicts (void)
{
#ifdef HAVE_LIBASPELL

  gtk_widget_destroy (mni_spellcheck_menu);
  mni_spellcheck_menu = new_menu_submenu (mni_spellcheck);
  GtkWidget *temp = new_menu_tof (mni_spellcheck_menu);

  GList *l = get_speller_modules_list ("ru");  
  GList *t = g_list_first (l);
  while (t)
        {
         mni_temp = new_menu_item (t->data, mni_spellcheck_menu, on_mni_spellcheck);
         t = g_list_next (t);
        }  
   
  glist_strings_free (l);

#endif
}


gchar* get_tmp_file_string (void)
{
   return str_file_read (confile.tmp_file); 
   
}


void build_user_menu (gpointer key, gpointer value, gpointer user_data)
{
  mni_temp = new_menu_item ((gchar*)key, mni_user_menu_menu, on_mni_user_menu_item);
}


void reload_usermenu (void)
{
  if (! confile.user_menu_file)
     return;

  if (! g_file_test (confile.user_menu_file, G_FILE_TEST_EXISTS))
     return;

  if (ht_user_menu)
     g_hash_table_destroy (ht_user_menu); 

  ht_user_menu = load_file_to_hashtable (confile.user_menu_file);
 
  gtk_widget_destroy (mni_user_menu_menu);
  mni_user_menu_menu = new_menu_submenu (mni_user_menu);
  mni_temp = new_menu_tof (mni_user_menu_menu);

  g_hash_table_foreach (ht_user_menu, build_user_menu, NULL); 
}
